@@ -1,23 +1,25 @@
 use super::helpers::{
     allocations,
-    edits::invert_edit,
-    edits::ReadRecorder,
-    fixtures::{get_language, get_test_grammar, get_test_language},
+    edits::{invert_edit, ReadRecorder},
+    fixtures::{get_language, get_test_language},
 };
 use crate::{
     generate::generate_parser_for_grammar,
     parse::{perform_edit, Edit},
+    tests::helpers::fixtures::fixtures_dir,
 };
 use std::{
+    fs,
     sync::atomic::{AtomicUsize, Ordering},
     thread, time,
 };
 use tree_sitter::{IncludedRangesError, InputEdit, LogType, Parser, Point, Range};
+use tree_sitter_proc_macro::retry;
 
 #[test]
 fn test_parsing_simple_string() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
 
     let tree = parser
         .parse(
@@ -48,7 +50,7 @@ fn test_parsing_simple_string() {
 #[test]
 fn test_parsing_with_logging() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
 
     let mut messages = Vec::new();
     parser.set_logger(Some(Box::new(|log_type, message| {
@@ -89,7 +91,7 @@ fn test_parsing_with_debug_graph_enabled() {
     let has_zero_indexed_row = |s: &str| s.contains("position: 0,");
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
 
     let mut debug_graph_file = tempfile::tempfile().unwrap();
     parser.print_dot_graphs(&debug_graph_file);
@@ -102,8 +104,7 @@ fn test_parsing_with_debug_graph_enabled() {
     for line in log_reader {
         assert!(
             !has_zero_indexed_row(&line),
-            "Graph log output includes zero-indexed row: {}",
-            line
+            "Graph log output includes zero-indexed row: {line}",
         );
     }
 }
@@ -111,7 +112,7 @@ fn test_parsing_with_debug_graph_enabled() {
 #[test]
 fn test_parsing_with_custom_utf8_input() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
 
     let lines = &["pub fn foo() {", "  1", "}"];
 
@@ -124,7 +125,7 @@ fn test_parsing_with_custom_utf8_input() {
                     if column < lines[row].as_bytes().len() {
                         &lines[row].as_bytes()[column..]
                     } else {
-                        "\n".as_bytes()
+                        b"\n"
                     }
                 } else {
                     &[]
@@ -147,19 +148,19 @@ fn test_parsing_with_custom_utf8_input() {
         )
     );
     assert_eq!(root.kind(), "source_file");
-    assert_eq!(root.has_error(), false);
+    assert!(!root.has_error());
     assert_eq!(root.child(0).unwrap().kind(), "function_item");
 }
 
 #[test]
 fn test_parsing_with_custom_utf16_input() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
 
-    let lines: Vec<Vec<u16>> = ["pub fn foo() {", "  1", "}"]
+    let lines = ["pub fn foo() {", "  1", "}"]
         .iter()
-        .map(|s| s.encode_utf16().collect())
-        .collect();
+        .map(|s| s.encode_utf16().collect::<Vec<_>>())
+        .collect::<Vec<_>>();
 
     let tree = parser
         .parse_utf16_with(
@@ -186,14 +187,14 @@ fn test_parsing_with_custom_utf16_input() {
         "(source_file (function_item (visibility_modifier) name: (identifier) parameters: (parameters) body: (block (integer_literal))))"
     );
     assert_eq!(root.kind(), "source_file");
-    assert_eq!(root.has_error(), false);
+    assert!(!root.has_error());
     assert_eq!(root.child(0).unwrap().kind(), "function_item");
 }
 
 #[test]
 fn test_parsing_with_callback_returning_owned_strings() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
 
     let text = b"pub fn foo() { 1 }";
 
@@ -214,7 +215,7 @@ fn test_parsing_with_callback_returning_owned_strings() {
 #[test]
 fn test_parsing_text_with_byte_order_mark() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
 
     // Parse UTF16 text with a BOM
     let tree = parser
@@ -273,15 +274,18 @@ fn test_parsing_text_with_byte_order_mark() {
 #[test]
 fn test_parsing_invalid_chars_at_eof() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("json")).unwrap();
+    parser.set_language(&get_language("json")).unwrap();
     let tree = parser.parse(b"\xdf", None).unwrap();
-    assert_eq!(tree.root_node().to_sexp(), "(ERROR (UNEXPECTED INVALID))");
+    assert_eq!(
+        tree.root_node().to_sexp(),
+        "(document (ERROR (UNEXPECTED INVALID)))"
+    );
 }
 
 #[test]
 fn test_parsing_unexpected_null_characters_within_source() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     let tree = parser.parse(b"var \0 something;", None).unwrap();
     assert_eq!(
         tree.root_node().to_sexp(),
@@ -292,7 +296,7 @@ fn test_parsing_unexpected_null_characters_within_source() {
 #[test]
 fn test_parsing_ends_when_input_callback_returns_empty() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     let mut i = 0;
     let source = b"abcdefghijklmnoqrs";
     let tree = parser
@@ -316,7 +320,7 @@ fn test_parsing_ends_when_input_callback_returns_empty() {
 #[test]
 fn test_parsing_after_editing_beginning_of_code() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
 
     let mut code = b"123 + 456 * (10 + x);".to_vec();
     let mut tree = parser.parse(&code, None).unwrap();
@@ -338,7 +342,8 @@ fn test_parsing_after_editing_beginning_of_code() {
             deleted_length: 0,
             inserted_text: b" || 5".to_vec(),
         },
-    );
+    )
+    .unwrap();
 
     let mut recorder = ReadRecorder::new(&code);
     let tree = parser
@@ -363,7 +368,7 @@ fn test_parsing_after_editing_beginning_of_code() {
 #[test]
 fn test_parsing_after_editing_end_of_code() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
 
     let mut code = b"x * (100 + abc);".to_vec();
     let mut tree = parser.parse(&code, None).unwrap();
@@ -385,7 +390,8 @@ fn test_parsing_after_editing_end_of_code() {
             deleted_length: 0,
             inserted_text: b".d".to_vec(),
         },
-    );
+    )
+    .unwrap();
 
     let mut recorder = ReadRecorder::new(&code);
     let tree = parser
@@ -410,7 +416,7 @@ fn test_parsing_after_editing_end_of_code() {
 #[test]
 fn test_parsing_empty_file_with_reused_tree() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
 
     let tree = parser.parse("", None);
     parser.parse("", tree.as_ref());
@@ -421,16 +427,15 @@ fn test_parsing_empty_file_with_reused_tree() {
 
 #[test]
 fn test_parsing_after_editing_tree_that_depends_on_column_values() {
-    let (grammar, path) = get_test_grammar("uses_current_column");
+    let dir = fixtures_dir()
+        .join("test_grammars")
+        .join("uses_current_column");
+    let grammar = fs::read_to_string(dir.join("grammar.json")).unwrap();
     let (grammar_name, parser_code) = generate_parser_for_grammar(&grammar).unwrap();
 
     let mut parser = Parser::new();
     parser
-        .set_language(get_test_language(
-            &grammar_name,
-            &parser_code,
-            path.as_ref().map(AsRef::as_ref),
-        ))
+        .set_language(&get_test_language(&grammar_name, &parser_code, Some(&dir)))
         .unwrap();
 
     let mut code = b"
@@ -460,7 +465,8 @@ h + i
             deleted_length: 0,
             inserted_text: b"1234".to_vec(),
         },
-    );
+    )
+    .unwrap();
 
     assert_eq!(
         code,
@@ -499,13 +505,13 @@ h + i
 #[test]
 fn test_parsing_after_detecting_error_in_the_middle_of_a_string_token() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("python")).unwrap();
+    parser.set_language(&get_language("python")).unwrap();
 
     let mut source = b"a = b, 'c, d'".to_vec();
     let tree = parser.parse(&source, None).unwrap();
     assert_eq!(
         tree.root_node().to_sexp(),
-        "(module (expression_statement (assignment left: (identifier) right: (expression_list (identifier) (string)))))"
+        "(module (expression_statement (assignment left: (identifier) right: (expression_list (identifier) (string (string_start) (string_content) (string_end))))))"
     );
 
     // Delete a suffix of the source code, starting in the middle of the string
@@ -524,12 +530,12 @@ fn test_parsing_after_detecting_error_in_the_middle_of_a_string_token() {
     let undo = invert_edit(&source, &edit);
 
     let mut tree2 = tree.clone();
-    perform_edit(&mut tree2, &mut source, &edit);
+    perform_edit(&mut tree2, &mut source, &edit).unwrap();
     tree2 = parser.parse(&source, Some(&tree2)).unwrap();
     assert!(tree2.root_node().has_error());
 
     let mut tree3 = tree2.clone();
-    perform_edit(&mut tree3, &mut source, &undo);
+    perform_edit(&mut tree3, &mut source, &undo).unwrap();
     tree3 = parser.parse(&source, Some(&tree3)).unwrap();
     assert_eq!(tree3.root_node().to_sexp(), tree.root_node().to_sexp(),);
 }
@@ -543,7 +549,7 @@ fn test_parsing_on_multiple_threads() {
     let this_file_source = include_str!("parser_test.rs");
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("rust")).unwrap();
+    parser.set_language(&get_language("rust")).unwrap();
     let tree = parser.parse(this_file_source, None).unwrap();
 
     let mut parse_threads = Vec::new();
@@ -571,7 +577,7 @@ fn test_parsing_on_multiple_threads() {
 
             // Reparse using the old tree as a starting point.
             let mut parser = Parser::new();
-            parser.set_language(get_language("rust")).unwrap();
+            parser.set_language(&get_language("rust")).unwrap();
             parser.parse(&prepended_source, Some(&tree_clone)).unwrap()
         }));
     }
@@ -592,7 +598,7 @@ fn test_parsing_cancelled_by_another_thread() {
     let cancellation_flag = std::sync::Arc::new(AtomicUsize::new(0));
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     unsafe { parser.set_cancellation_flag(Some(&cancellation_flag)) };
 
     // Long input - parsing succeeds
@@ -638,9 +644,10 @@ fn test_parsing_cancelled_by_another_thread() {
 // Timeouts
 
 #[test]
+#[retry(10)]
 fn test_parsing_with_a_timeout() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("json")).unwrap();
+    parser.set_language(&get_language("json")).unwrap();
 
     // Parse an infinitely-long array, but pause after 1ms of processing.
     parser.set_timeout_micros(1000);
@@ -679,14 +686,10 @@ fn test_parsing_with_a_timeout() {
     parser.set_timeout_micros(0);
     let tree = parser
         .parse_with(
-            &mut |offset, _| {
-                if offset > 5000 {
-                    "".as_bytes()
-                } else if offset == 5000 {
-                    "]".as_bytes()
-                } else {
-                    ",0".as_bytes()
-                }
+            &mut |offset, _| match offset {
+                5001.. => "".as_bytes(),
+                5000 => "]".as_bytes(),
+                _ => ",0".as_bytes(),
             },
             None,
         )
@@ -695,9 +698,10 @@ fn test_parsing_with_a_timeout() {
 }
 
 #[test]
+#[retry(10)]
 fn test_parsing_with_a_timeout_and_a_reset() {
     let mut parser = Parser::new();
-    parser.set_language(get_language("json")).unwrap();
+    parser.set_language(&get_language("json")).unwrap();
 
     parser.set_timeout_micros(5);
     let tree = parser.parse(
@@ -750,10 +754,11 @@ fn test_parsing_with_a_timeout_and_a_reset() {
 }
 
 #[test]
+#[retry(10)]
 fn test_parsing_with_a_timeout_and_implicit_reset() {
     allocations::record(|| {
         let mut parser = Parser::new();
-        parser.set_language(get_language("javascript")).unwrap();
+        parser.set_language(&get_language("javascript")).unwrap();
 
         parser.set_timeout_micros(5);
         let tree = parser.parse(
@@ -764,7 +769,7 @@ fn test_parsing_with_a_timeout_and_implicit_reset() {
 
         // Changing the parser's language implicitly resets, discarding
         // the previous partial parse.
-        parser.set_language(get_language("json")).unwrap();
+        parser.set_language(&get_language("json")).unwrap();
         parser.set_timeout_micros(0);
         let tree = parser.parse(
             "[null, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32]",
@@ -783,10 +788,11 @@ fn test_parsing_with_a_timeout_and_implicit_reset() {
 }
 
 #[test]
+#[retry(10)]
 fn test_parsing_with_timeout_and_no_completion() {
     allocations::record(|| {
         let mut parser = Parser::new();
-        parser.set_language(get_language("javascript")).unwrap();
+        parser.set_language(&get_language("javascript")).unwrap();
 
         parser.set_timeout_micros(5);
         let tree = parser.parse(
@@ -806,7 +812,7 @@ fn test_parsing_with_one_included_range() {
     let source_code = "<span>hi</span><script>console.log('sup');</script>";
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("html")).unwrap();
+    parser.set_language(&get_language("html")).unwrap();
     let html_tree = parser.parse(source_code, None).unwrap();
     let script_content_node = html_tree.root_node().child(1).unwrap().child(1).unwrap();
     assert_eq!(script_content_node.kind(), "raw_text");
@@ -814,7 +820,7 @@ fn test_parsing_with_one_included_range() {
     parser
         .set_included_ranges(&[script_content_node.range()])
         .unwrap();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     let js_tree = parser.parse(source_code, None).unwrap();
 
     assert_eq!(
@@ -822,7 +828,7 @@ fn test_parsing_with_one_included_range() {
         concat!(
             "(program (expression_statement (call_expression ",
             "function: (member_expression object: (identifier) property: (property_identifier)) ",
-            "arguments: (arguments (string)))))",
+            "arguments: (arguments (string (string_fragment))))))",
         )
     );
     assert_eq!(
@@ -837,23 +843,23 @@ fn test_parsing_with_multiple_included_ranges() {
     let source_code = "html `<div>Hello, ${name.toUpperCase()}, it's <b>${now()}</b>.</div>`";
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     let js_tree = parser.parse(source_code, None).unwrap();
     let template_string_node = js_tree
         .root_node()
         .descendant_for_byte_range(
-            source_code.find("<div>").unwrap(),
-            source_code.find("Hello").unwrap(),
+            source_code.find("`<").unwrap(),
+            source_code.find(">`").unwrap(),
         )
         .unwrap();
     assert_eq!(template_string_node.kind(), "template_string");
 
     let open_quote_node = template_string_node.child(0).unwrap();
-    let interpolation_node1 = template_string_node.child(1).unwrap();
-    let interpolation_node2 = template_string_node.child(2).unwrap();
-    let close_quote_node = template_string_node.child(3).unwrap();
+    let interpolation_node1 = template_string_node.child(2).unwrap();
+    let interpolation_node2 = template_string_node.child(4).unwrap();
+    let close_quote_node = template_string_node.child(6).unwrap();
 
-    parser.set_language(get_language("html")).unwrap();
+    parser.set_language(&get_language("html")).unwrap();
     let html_ranges = &[
         Range {
             start_byte: open_quote_node.end_byte(),
@@ -880,7 +886,7 @@ fn test_parsing_with_multiple_included_ranges() {
     assert_eq!(
         html_tree.root_node().to_sexp(),
         concat!(
-            "(fragment (element",
+            "(document (element",
             " (start_tag (tag_name))",
             " (text)",
             " (element (start_tag (tag_name)) (end_tag (tag_name)))",
@@ -932,7 +938,7 @@ fn test_parsing_with_included_range_containing_mismatched_positions() {
     let source_code = "<div>test</div>{_ignore_this_part_}";
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("html")).unwrap();
+    parser.set_language(&get_language("html")).unwrap();
 
     let end_byte = source_code.find("{_ignore_this_part_").unwrap();
 
@@ -959,7 +965,7 @@ fn test_parsing_with_included_range_containing_mismatched_positions() {
 
     assert_eq!(
         html_tree.root_node().to_sexp(),
-        "(fragment (element (start_tag (tag_name)) (text) (end_tag (tag_name))))"
+        "(document (element (start_tag (tag_name)) (text) (end_tag (tag_name))))"
     );
 }
 
@@ -1007,13 +1013,17 @@ fn test_parsing_error_in_invalid_included_ranges() {
 #[test]
 fn test_parsing_utf16_code_with_errors_at_the_end_of_an_included_range() {
     let source_code = "<script>a.</script>";
-    let utf16_source_code: Vec<u16> = source_code.as_bytes().iter().map(|c| *c as u16).collect();
+    let utf16_source_code = source_code
+        .as_bytes()
+        .iter()
+        .map(|c| u16::from(*c))
+        .collect::<Vec<_>>();
 
     let start_byte = 2 * source_code.find("a.").unwrap();
     let end_byte = 2 * source_code.find("</script>").unwrap();
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     parser
         .set_included_ranges(&[Range {
             start_byte,
@@ -1035,7 +1045,7 @@ fn test_parsing_with_external_scanner_that_uses_included_range_boundaries() {
     let range2_end_byte = range2_start_byte + " d() ".len();
 
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     parser
         .set_included_ranges(&[
             Range {
@@ -1079,7 +1089,7 @@ fn test_parsing_with_a_newly_excluded_range() {
 
     // Parse HTML including the template directive, which will cause an error
     let mut parser = Parser::new();
-    parser.set_language(get_language("html")).unwrap();
+    parser.set_language(&get_language("html")).unwrap();
     let mut first_tree = parser
         .parse_with(&mut chunked_input(&source_code, 3), None)
         .unwrap();
@@ -1124,7 +1134,7 @@ fn test_parsing_with_a_newly_excluded_range() {
     assert_eq!(
         tree.root_node().to_sexp(),
         concat!(
-            "(fragment (text) (element",
+            "(document (text) (element",
             " (start_tag (tag_name))",
             " (element (start_tag (tag_name)) (end_tag (tag_name)))",
             " (end_tag (tag_name))))"
@@ -1166,12 +1176,12 @@ fn test_parsing_with_a_newly_included_range() {
 
     // Parse only the first code directive as JavaScript
     let mut parser = Parser::new();
-    parser.set_language(get_language("javascript")).unwrap();
+    parser.set_language(&get_language("javascript")).unwrap();
     parser
         .set_included_ranges(&[simple_range(range1_start, range1_end)])
         .unwrap();
     let tree = parser
-        .parse_with(&mut chunked_input(&source_code, 3), None)
+        .parse_with(&mut chunked_input(source_code, 3), None)
         .unwrap();
     assert_eq!(
         tree.root_node().to_sexp(),
@@ -1190,7 +1200,7 @@ fn test_parsing_with_a_newly_included_range() {
         ])
         .unwrap();
     let tree2 = parser
-        .parse_with(&mut chunked_input(&source_code, 3), Some(&tree))
+        .parse_with(&mut chunked_input(source_code, 3), Some(&tree))
         .unwrap();
     assert_eq!(
         tree2.root_node().to_sexp(),
@@ -1214,7 +1224,7 @@ fn test_parsing_with_a_newly_included_range() {
             simple_range(range3_start, range3_end),
         ])
         .unwrap();
-    let tree3 = parser.parse(&source_code, Some(&tree)).unwrap();
+    let tree3 = parser.parse(source_code, Some(&tree)).unwrap();
     assert_eq!(
         tree3.root_node().to_sexp(),
         concat!(
@@ -1258,7 +1268,7 @@ fn test_parsing_with_included_ranges_and_missing_tokens() {
 
     let mut parser = Parser::new();
     parser
-        .set_language(get_test_language(&parser_name, &parser_code, None))
+        .set_language(&get_test_language(&parser_name, &parser_code, None))
         .unwrap();
 
     // There's a missing `a` token at the beginning of the code. It must be inserted
@@ -1291,7 +1301,119 @@ fn test_parsing_with_included_ranges_and_missing_tokens() {
     assert_eq!(root.child(3).unwrap().start_byte(), 4);
 }
 
-fn simple_range(start: usize, end: usize) -> Range {
+#[test]
+fn test_grammars_that_can_hang_on_eof() {
+    let (parser_name, parser_code) = generate_parser_for_grammar(
+        r#"
+        {
+            "name": "test_single_null_char_regex",
+            "rules": {
+                "source_file": {
+                    "type": "SEQ",
+                    "members": [
+                        { "type": "STRING", "value": "\"" },
+                        { "type": "PATTERN", "value": "[\\x00]*" },
+                        { "type": "STRING", "value": "\"" }
+                    ]
+                }
+            },
+            "extras": [ { "type": "PATTERN", "value": "\\s" } ]
+        }
+        "#,
+    )
+    .unwrap();
+
+    let mut parser = Parser::new();
+    parser
+        .set_language(&get_test_language(&parser_name, &parser_code, None))
+        .unwrap();
+    parser.parse("\"", None).unwrap();
+
+    let (parser_name, parser_code) = generate_parser_for_grammar(
+        r#"
+        {
+            "name": "test_null_char_with_next_char_regex",
+            "rules": {
+                "source_file": {
+                    "type": "SEQ",
+                    "members": [
+                        { "type": "STRING", "value": "\"" },
+                        { "type": "PATTERN", "value": "[\\x00-\\x01]*" },
+                        { "type": "STRING", "value": "\"" }
+                    ]
+                }
+            },
+            "extras": [ { "type": "PATTERN", "value": "\\s" } ]
+        }
+        "#,
+    )
+    .unwrap();
+
+    parser
+        .set_language(&get_test_language(&parser_name, &parser_code, None))
+        .unwrap();
+    parser.parse("\"", None).unwrap();
+
+    let (parser_name, parser_code) = generate_parser_for_grammar(
+        r#"
+        {
+            "name": "test_null_char_with_range_regex",
+            "rules": {
+                "source_file": {
+                    "type": "SEQ",
+                    "members": [
+                        { "type": "STRING", "value": "\"" },
+                        { "type": "PATTERN", "value": "[\\x00-\\x7F]*" },
+                        { "type": "STRING", "value": "\"" }
+                    ]
+                }
+            },
+            "extras": [ { "type": "PATTERN", "value": "\\s" } ]
+        }
+        "#,
+    )
+    .unwrap();
+
+    parser
+        .set_language(&get_test_language(&parser_name, &parser_code, None))
+        .unwrap();
+    parser.parse("\"", None).unwrap();
+}
+
+#[test]
+fn test_parse_stack_recursive_merge_error_cost_calculation_bug() {
+    let source_code = r#"
+fn main() {
+  if n == 1 {
+  } else if n == 2 {
+  } else {
+  }
+}
+
+let y = if x == 5 { 10 } else { 15 };
+
+if foo && bar {}
+
+if foo && bar || baz {}
+"#;
+
+    let mut parser = Parser::new();
+    parser.set_language(&get_language("rust")).unwrap();
+
+    let mut tree = parser.parse(source_code, None).unwrap();
+
+    let edit = Edit {
+        position: 60,
+        deleted_length: 63,
+        inserted_text: Vec::new(),
+    };
+    let mut input = source_code.as_bytes().to_vec();
+    perform_edit(&mut tree, &mut input, &edit).unwrap();
+
+    parser.parse(&input, Some(&tree)).unwrap();
+}
+
+const fn simple_range(start: usize, end: usize) -> Range {
     Range {
         start_byte: start,
         end_byte: end,
