@@ -22,7 +22,7 @@ fn test_tags_test_with_basic_test() {
     .join("\n");
 
     let assertions =
-        parse_position_comments(&mut Parser::new(), language, source.as_bytes()).unwrap();
+        parse_position_comments(&mut Parser::new(), &language, source.as_bytes()).unwrap();
 
     assert_eq!(
         assertions,
@@ -55,5 +55,5 @@ fn test_tags_test_with_basic_test() {
                 "reference.call".to_string()
             ),
         ]
-    )
+    );
 }
